@@ -2,6 +2,7 @@ package cap
 
 import (
 	"errors"
+	"fmt"
 	"syscall"
 	"unsafe"
 )
@@ -33,9 +34,11 @@ const (
 
 // defines from uapi/linux/prctl.h
 const (
+	prGetKeepCaps   = 7
 	prSetKeepCaps   = 8
 	prGetSecureBits = 27
 	prSetSecureBits = 28
+	prSetNoNewPrivs = 38
 )
 
 // GetSecbits returns the current setting of the process' Secbits.
@@ -56,9 +59,9 @@ func (sc *syscaller) setSecbits(s Secbits) error {
 // will raise cap.SETPCAP in order to achieve this operation, and will
 // completely lower the Effective  vector of the process returning.
 func (s Secbits) Set() error {
-	scwMu.Lock()
-	defer scwMu.Unlock()
-	return multisc.setSecbits(s)
+	state, sc := scwStateSC()
+	defer scwSetState(launchBlocked, state, -1)
+	return sc.setSecbits(s)
 }
 
 // Mode summarizes a complicated secure-bits and capability mode in a
@@ -163,6 +166,9 @@ func (sc *syscaller) setMode(m Mode) error {
 	}
 	w.ClearFlag(Permitted)
 
+	// For good measure.
+	sc.prctlwcall6(prSetNoNewPrivs, 1, 0, 0, 0, 0)
+
 	return nil
 }
 
@@ -172,14 +178,14 @@ func (sc *syscaller) setMode(m Mode) error {
 // the desired mode.
 //
 // This function will raise cap.SETPCAP in order to achieve this
-// operation, and will completely lower the Effective dimension of the
+// operation, and will completely lower the Effective Flag of the
 // process' Set before returning. This function may fail for lack of
 // permission or because (some of) the Secbits are already locked for
 // the current process.
 func (m Mode) Set() error {
-	scwMu.Lock()
-	defer scwMu.Unlock()
-	return multisc.setMode(m)
+	state, sc := scwStateSC()
+	defer scwSetState(launchBlocked, state, -1)
+	return sc.setMode(m)
 }
 
 // String returns the libcap conventional string for this mode.
@@ -229,14 +235,14 @@ func (sc *syscaller) setUID(uid int) error {
 // dropping the privilege of the current process. This function will
 // raise cap.SETUID in order to achieve this operation, and will
 // completely lower the Effective vector of the process before
-// returning. Unlike the traditional method of dropping privilege
-// when changing from [E]UID=0 to some other UID, this function only
+// returning. Unlike the traditional method of dropping privilege when
+// changing from [E]UID=0 to some other UID, this function only
 // performs a change of UID cap.SETUID is available, and the action
-// does not alter the Permitted dimension of the process' Set.
+// does not alter the Permitted Flag of the process' Set.
 func SetUID(uid int) error {
-	scwMu.Lock()
-	defer scwMu.Unlock()
-	return multisc.setUID(uid)
+	state, sc := scwStateSC()
+	defer scwSetState(launchBlocked, state, -1)
+	return sc.setUID(uid)
 }
 
 //go:uintptrescapes
@@ -279,10 +285,46 @@ func (sc *syscaller) setGroups(gid int, suppl []int) error {
 // and all other variants of GID (EGID etc) to the specified value, as
 // well as setting all of the supplementary groups. This function will
 // raise cap.SETGID in order to achieve this operation, and will
-// completely lower the Effective dimension of the process Set before
+// completely lower the Effective Flag of the process Set before
 // returning.
 func SetGroups(gid int, suppl ...int) error {
-	scwMu.Lock()
-	defer scwMu.Unlock()
-	return multisc.setGroups(gid, suppl)
+	state, sc := scwStateSC()
+	defer scwSetState(launchBlocked, state, -1)
+	return sc.setGroups(gid, suppl)
+}
+
+//go:uintptrescapes
+
+// Prctlw is a convenience function for performing a syscall.Prctl()
+// call that executes on all the threads of the process. It is called
+// Prctlw because it is only appropriate to call this function when it
+// is writing thread state that the caller wants to set on all OS
+// threads of the process to observe POSIX semantics when Linux
+// doesn't natively honor them. (Check prctl documentation for when it
+// is appropriate to use this vs. a normal syscall.Prctl() call.)
+func Prctlw(prVal uintptr, args ...uintptr) (int, error) {
+	if n := len(args); n > 5 {
+		return -1, fmt.Errorf("prctl supports up to 5 arguments (not %d)", n)
+	}
+	state, sc := scwStateSC()
+	defer scwSetState(launchBlocked, state, -1)
+	as := make([]uintptr, 5)
+	copy(as, args)
+	return sc.prctlwcall6(prVal, as[0], as[1], as[2], as[3], as[4])
+}
+
+//go:uintptrescapes
+
+// Prctl is a convenience function that performs a syscall.Prctl()
+// that either reads state using a single OS thread, or performs a
+// Prctl that is treated as a process wide setting. It is provided for
+// symmetry reasons, but is equivalent to simply calling the
+// corresponding syscall function.
+func Prctl(prVal uintptr, args ...uintptr) (int, error) {
+	if n := len(args); n > 5 {
+		return -1, fmt.Errorf("prctl supports up to 5 arguments (not %d)", n)
+	}
+	as := make([]uintptr, 5)
+	copy(as, args)
+	return singlesc.prctlrcall6(prVal, as[0], as[1], as[2], as[3], as[4])
 }
